---
title: 自定义事件
date: 2020-02-15 11:27:45
permalink: /pages/9651417d08d1779d
categories:
  - 《Vue》笔记
  - 组件
tags:
  - Vue
author:
  name: xugaoyi
  link: https://github.com/xugaoyi
---
# 自定义事件

[API](https://cn.vuejs.org/v2/guide/components-custom-events.html)

### 在组件标签v-on绑定的事件是自定义事件

```html
<div id="root">
    <child @click="handleClick"></child> <!--这里click是自定义事件-->
</div>

<script type="text/javascript">
    Vue.component('child', {
        template: '<button>Child</button>',
    })

    var vm = new Vue({
        el: '#root'
        methods: {
            handleClick() {
                alert('click')
            }
        }
    })
</script>
```

上面这种情况，当点击子组件时并不会触发`handleClick()`事件，因为此时的`click`并非原生的点击事件，而是一个自定义事件。



要想触发组件标签中的`click`事件需要通过子组件`$emit`来派发

```html
<div id="root">
    <child @click="handleClick"></child>
</div>

<script type="text/javascript">
    Vue.component('child', {
        template: '<button @click="handleChildClick">Child</button>', // 这里的click是原生事件
        methods: {
            handleChildClick() {
                this.$emit('click')
            }
        }
    })

    var vm = new Vue({
        el: '#root'
        methods: {
            handleClick() {
                alert('click')
            }
        }
    })
</script>
```

上面代码中，子组件内的`click`是原生点击事件，通过`$emit`派发`click`自定义事件触发父组件的`handleClick()`



### 使自定义事件变为原生事件

使用`.native`修饰符把自定义事件变为原生事件

```html
<div id="root">
    <child @click.native="handleClick"></child> <!--自定义事件添加了native修饰符变原生事件-->
</div>

<script type="text/javascript">
    Vue.component('child', {
        template: '<button>Child</button>',
    })

    var vm = new Vue({
        el: '#root'
        methods: {
            handleClick() {
                alert('click')
            }
        }
    })
</script>
```

上面代码中，自定义事件`click`添加了`.native`修饰符，使其变为原生事件，从而触发`handleClick()`。



### 自定义事件名会自动变小写

当使用camelCase（驼峰）命名的自定义函数

```js
this.$emit('myEvent')
```

则监听这个名字的 kebab-case（烤串） 版本是不会有任何效果的：

```html
<!-- 没有效果 -->
<my-component v-on:my-event="doSomething"></my-component>
```

 `v-on` 事件监听器在 DOM 模板中会被自动转换为全小写 (因为 HTML 是大小写不敏感的)，所以 `v-on:myEvent` 将会变成 `v-on:myevent`——导致 `myEvent` 不可能被监听到。

因此，我们推荐你**始终使用 kebab-case（烤串） 的事件名**。
